import spacy, textacy, pyvis
import pandas as pd
from textblob import TextBlob
from tqdm import tqdm

import warnings
warnings.filterwarnings('ignore')

def getlabel(x):
  return [y.label_ for y in x.ents] \
    if [y.label_ for y in x.ents] else None
    

def SVO(text, sentiment=False): # return SVO table from given text
  try:
    nlp = spacy.load('en_core_web_sm')
  except:
    print('Install spacy model first by running\npython -m spacy download en_core_web_sm')
    return
  if not text: 
    print('no text found') # empty text warning
    return
  doc = nlp(text) 
  print('found ',len(list(doc.sents)), 'sentences in the corpus, extracting items')
  out = [] # prepare an empty list to receive result
  with tqdm(total=(len(list(doc.sents)))) as tm:
    for sent in doc.sents: # loop each sentence in the text
      out = []
      for sent in doc.sents:
        elements = list(textacy.extract.subject_verb_object_triples(sent))
        if elements:
          for element in elements:
            element = {'Subject':element[0].text,'SubjectType': getlabel(element[0]),
                      'Verb':element[1].text, 'Object':element[2].text, 'ObjectType':getlabel(element[2])}
            if sentiment:
              score = TextBlob(' '.join([d.text for d in sent])).sentiment.polarity
              element['sentiment'] = score
            out.append(element)
      tm.update()
  return pd.DataFrame(out)

def SVOold(text, sentiment=False): # return SVO table from given text
  try:
    nlp = spacy.load('en_core_web_sm')
  except:
    print('Install spacy model first by running\npython -m spacy download en_core_web_sm')
    return
  if not text: 
    print('no text found') # empty text warning
    return
  doc = nlp(text) 
  print('found ',len(list(doc.sents)), 'sentences in the corpus, extracting items')
  out = [] # prepare an empty list to receive result
  with tqdm(total=(len(list(doc.sents)))) as tm:
    for sent in doc.sents: # loop each sentence in the text
      if len(list(sent.ents))<0:
        pass
        tm.update()
      elements = list(textacy.extract.subject_verb_object_triples(sent))
      # generate svo list
      if sentiment: # check whether need return sentiments
        score = TextBlob(' '.join([d.text for d in sent])).sentiment.polarity
        # return sentment score for the sentencezt
        elements = [(e[0],e[1],e[2], score) for e in elements]
        # reoganize the sentence analysis result
      out += elements # add sentence result to the output list 
      tm.update()
  columns=['Subject','Verb','Object'] # define the structure of output table
  if sentiment:
    columns.append('Sentiment') # add sentiment column if selected
  svodf= pd.DataFrame(out, columns=columns) # sent result to dataframe
  svodf['SubjectType'] =svodf.Subject.apply(lambda x: [y.label_ for y in x.ents] \
    if [y.label_ for y in x.ents] else None)
  svodf['ObjectType'] =svodf.Object.apply(lambda x: [y.label_ for y in x.ents] \
    if [y.label_ for y in x.ents] else None)
  svodf.Subject =svodf.Subject.apply(lambda x: x.text)
  svodf.Object =svodf.Object.apply(lambda x: x.text)
  svodf.Verb =svodf.Verb.apply(lambda x: x.text)
  newCols = ['Subject', 'SubjectType', 'Verb', 'Object','ObjectType']
  if sentiment:
    newCols.append('Sentiment') # add sentiment column if selected
  svodf= svodf[newCols]
  return svodf

def visSVO(svodf, filename='', options='entity'): # option could be "entity", "person", "any"
  from IPython.display import HTML, display
  filename = f'network_{filename}.html'
  print('creating ', filename, '...')
  if len(svodf) <1:
    print('no entity in the dataframe, try another one')
    return 
  if options in ["entity","person"]: # check options
    svodf=svodf.dropna(subset=['SubjectType']).reset_index(drop=True) # remove non Type subjects
    if options == "person": # check options
      svodf=svodf.loc[svodf.SubjectType.apply(lambda x: 'PERSON' in x),:] 
      # select only person as subject relations
  nodeS = list(set(svodf.Subject)) # create subject nodes
  nodeO = list(set(svodf.Object)) # create object nodes
  nodeAll = list(set(nodeS+nodeO)) # merge and uniq all nodes
  nodes = range(len(nodeAll)) # create node for the visualization (must be numbers)
  colors = ['maroon' if n in nodeS else 'TEAL' for n in nodeAll] 
  # set subject at color "maroon" and object as "teal" if both as "maroon"
  net = pyvis.network.Network(height='500px', width='1000px',) 
  # create a empty chart net
  net.add_nodes(nodes, label=nodeAll, color=colors, title=nodeAll) 
  # add nodes to network graph
  for i in svodf.index: # add relationships(Verbs) to the graph
    net.add_edge(nodeAll.index(svodf.loc[i,'Subject']),nodeAll.index(svodf.loc[i,'Object']),
     label=svodf.loc[i,'Verb'], title=str(i))
  net.show(filename) # save the graph to file "network.html"
  display(HTML(filename))
  return filename

def SVOall(df, textCol='KeyTopic'):
  topicSVOs = {}
  for topic in df[textCol].unique():
    text = '\n'.join(df.loc[df[textCol]==topic,'content'])
    topicSVO = SVO(text)
    # topicSVO.to_csv(f'svo_{topic}.csv', index=0)
    topicSVOs[topic] = topicSVO
  return topicSVOs



from gensim import corpora
from gensim.models import LdaModel
import pandas as pd
import re

from bokeh.plotting import figure, show, output_notebook
from bokeh.models import ColumnDataSource
from bokeh.palettes import Category20
from bokeh.transform import factor_cmap

import warnings

warnings.filterwarnings('ignore')

def topic_analysis(df, nTopics=5, cleanTextCol='cleaned_text'):
  cleandata = df[cleanTextCol].apply(lambda x: x.split(' '))
  dictionary = corpora.Dictionary(cleandata)
  tokens = [dictionary.doc2bow(d) for d in cleandata]
  model = LdaModel(tokens, num_topics=nTopics, id2word=dictionary, 
                update_every=1, chunksize=50, passes=10,
                per_word_topics=True, alpha='auto')
  docweights = [model.get_document_topics(t, minimum_probability=0) for t in tokens]
  doctopics = pd.DataFrame(docweights).apply(lambda x: x.apply(lambda y: y[-1] if y else 0))
  doctopics.columns = [f'topic{n+1}' for n in doctopics.columns]
  doctopics['KeyTopic']=doctopics.apply(lambda y:doctopics.columns[y==y.max()][0], axis=1)

  # create topicdescribe
  topics = model.show_topics(num_words=6)
  keywords = [re.findall(r'\*"(.*?)"',d[1]) for d in topics]
  weights = [re.findall(r'([\d\.]+)\*', d[1]) for d in topics]
  kwdf= pd.DataFrame(keywords, columns=[f'keyword_{n}' for n in range(len(keywords[0]))])
  wtdf= pd.DataFrame(weights, columns=[f'weight_{n}' for n in range(len(weights[0]))])
  topicDescribe =  kwdf.merge(wtdf,left_index=True, right_index=True)
  topicDescribe[sorted(topicDescribe.columns, key=lambda x:x.split('_')[-1])]
  topicDescribe['KeyTopic'] = [f'topic{n+1}' for n in range(len(topics))]
  topicDescribe['TopicKeywords'] = [' '.join(k) for k in keywords]
  topicDescribe['DocCount'] = doctopics['KeyTopic'].value_counts().sort_index().values
  topicDescribe = topicDescribe[['KeyTopic']+[col for col in topicDescribe.columns if \
    col != 'KeyTopic']]
  
  doctopics= doctopics.merge(topicDescribe[['KeyTopic','TopicKeywords']], on='KeyTopic', how='left')
  return doctopics, topicDescribe, model, tokens, dictionary


def getldaVis(model, tokens):
  import pyLDAvis.gensim
  try:
    pyLDAvis.enable_notebook()
  except:
    print('LDAvisualization should run in Jupyter notebook or Google colab')
  return pyLDAvis.gensim.prepare(model, tokens, dictionary=model.id2word)

def plot_topics(topicDescribe):
  try:
    output_notebook()
  except:
    print('LDAvisualization should run in Jupyter notebook or Google colab')
  source = ColumnDataSource(topicDescribe)
  tooltips = [('KeyTopic','@KeyTopic'),('keywords','@TopicKeywords'),('DocCount','@DocCount')]
  p = figure(x_range=list(topicDescribe['KeyTopic']), plot_height=400, plot_width=800,
            title="Topic Counts", toolbar_location=None, 
            tools="pan,wheel_zoom,box_zoom,reset,hover, save", 
            tooltips=tooltips)
  p.vbar(x = 'KeyTopic', top='DocCount',width=0.9, source=source,
   color=factor_cmap('KeyTopic', palette=Category20[20], factors=list(topicDescribe['KeyTopic'])))
  show(p)



#pip install IPython bokeh pandas xlrd gensim spacy textacy textblob pyvis pyLDAvis
# python -m spacy download en_core_web_sm
import pyLDAvis.gensim
import spacy, re,textacy,pyvis, IPython
import pandas as pd
from gensim import corpora
from gensim.models import LdaModel
from bokeh.plotting import figure, show, output_notebook
from bokeh.models import ColumnDataSource
from bokeh.palettes import Category20
from bokeh.transform import factor_cmap
import warnings
from textblob import TextBlob
warnings.filterwarnings('ignore')
# nlp = spacy.load('en_core_web_sm')
try:
  nlp = spacy.load('en_core_web_sm')
except:
  print('Install spacy model first by running\npython -m spacy download en_core_web_sm')
pyLDAvis.enable_notebook()


#Step 1 add sentiment column
def sentimentSocre(text):
  blob = TextBlob(text)
  return blob.sentiment.polarity

# Step 2 text clean
def text_clean(text, custom_stopwrods=[], toLower=True):
  if toLower:
    text=text.lower()
  doc = nlp(text)
  return ' '.join([d.lemma_ for d in doc if \
    all([2<len(d.text)<20,d.is_alpha, not d.is_stop, d.lemma_ not in custom_stopwrods])])

# Step3 Topic Analysis

def topic_analysis(cleanCol, nTopics=5):
  cleandata = df['cleaned_text'].apply(lambda x: x.split(' '))
  dictionary = corpora.Dictionary(cleandata)
  tokens = [dictionary.doc2bow(d) for d in cleandata]
  model = LdaModel(tokens, num_topics=nTopics, id2word=dictionary, 
                update_every=1, chunksize=50, passes=10,
                per_word_topics=True, alpha='auto')
  docweights = [model.get_document_topics(t, minimum_probability=0) for t in tokens]
  doctopics = pd.DataFrame(docweights).apply(lambda x: x.apply(lambda y: y[-1] if y else 0))
  doctopics.columns = [f'topic{n+1}' for n in doctopics.columns]
  doctopics['KeyTopic']=doctopics.apply(lambda y:doctopics.columns[y==y.max()][0], axis=1)

  # create topicdescribe
  topics = model.show_topics(num_words=6)
  keywords = [re.findall(r'\*"(.*?)"',d[1]) for d in topics]
  weights = [re.findall(r'([\d\.]+)\*', d[1]) for d in topics]
  kwdf= pd.DataFrame(keywords, columns=[f'keyword_{n}' for n in range(len(keywords[0]))])
  wtdf= pd.DataFrame(weights, columns=[f'weight_{n}' for n in range(len(weights[0]))])
  topicDescribe =  kwdf.merge(wtdf,left_index=True, right_index=True)
  topicDescribe[sorted(topicDescribe.columns, key=lambda x:x.split('_')[-1])]
  topicDescribe['KeyTopic'] = [f'topic{n+1}' for n in range(len(topics))]
  topicDescribe['TopicKeywords'] = [' '.join(k) for k in keywords]
  topicDescribe['DocCount'] = doctopics['KeyTopic'].value_counts().sort_index().values
  topicDescribe = topicDescribe[['KeyTopic']+[col for col in topicDescribe.columns if \
    col != 'KeyTopic']]
  ldaVis = pyLDAvis.gensim.prepare(model, tokens, dictionary=model.id2word)
  doctopics= doctopics.merge(topicDescribe[['KeyTopic','TopicKeywords']], on='KeyTopic', how='left')
  return doctopics, topicDescribe, ldaVis

# Step 3.2 
def plot_topics(topicDescribe):
  output_notebook()
  source = ColumnDataSource(topicDescribe)
  tooltips = [('KeyTopic','@KeyTopic'),('keywords','@TopicKeywords'),('DocCount','@DocCount')]
  p = figure(x_range=list(topicDescribe['KeyTopic']), plot_height=400, plot_width=800,
            title="Topic Counts", toolbar_location=None, 
            tools="pan,wheel_zoom,box_zoom,reset,hover, save", 
            tooltips=tooltips)
  p.vbar(x = 'KeyTopic', top='DocCount',width=0.9, source=source,
   color=factor_cmap('KeyTopic', palette=Category20[20], factors=list(topicDescribe['KeyTopic'])))
  show(p)

  plot_topics(topicDescribe)

# Step 4 SVO
def SVO(text, sentiment=False): # return SVO table from given text
  if not text: 
    print('no text found') # empty text warning
    return
  doc = nlp(text) 
  out = [] # prepare an empty list to receive result
  for sent in doc.sents: # loop each sentence in the text
    '''
    todo filter sent by ent
    '''
    elements = list(textacy.extract.subject_verb_object_triples(sent))
    # generate svo list
    if sentiment: # check whether need return sentiments
      score = TextBlob(' '.join([d.text for d in sent])).sentiment.polarity
      # return sentment score for the sentence
      elements = [(e[0],e[1],e[2], score) for e in elements]
      # reoganize the sentence analysis result
    out += elements # add sentence result to the output list 
  columns=['Subject','Verb','Object'] # define the structure of output table
  if sentiment:
    columns.append('Sentiment') # add sentiment column if selected
  svodf= pd.DataFrame(out, columns=columns) # sent result to dataframe
  svodf['SubjectType'] =svodf.Subject.apply(lambda x: [y.label_ for y in x.ents] \
    if [y.label_ for y in x.ents] else None)
  svodf['ObjectType'] =svodf.Object.apply(lambda x: [y.label_ for y in x.ents] \
    if [y.label_ for y in x.ents] else None)
  svodf.Subject =svodf.Subject.apply(lambda x: x.text)
  svodf.Object =svodf.Object.apply(lambda x: x.text)
  svodf.Verb =svodf.Verb.apply(lambda x: x.text)
  newCols = ['Subject', 'SubjectType', 'Verb', 'Object','ObjectType']
  if sentiment:
    newCols.append('Sentiment') # add sentiment column if selected
  svodf= svodf[newCols]
  return svodf

def visSVO(svodf, filename='', options='entity'): # option could be "entity", "person", "any"
  filename = f'network_{filename}.html'
  if options in ["entity","person"]: # check options
    svodf=svodf.dropna(subset=['SubjectType']).reset_index(drop=True) # remove non Type subjects
    if options == "person": # check options
      svodf=svodf.loc[svodf.SubjectType.apply(lambda x: 'PERSON' in x),:] 
      # select only person as subject relations
  nodeS = list(set(svodf.Subject)) # create subject nodes
  nodeO = list(set(svodf.Object)) # create object nodes
  nodeAll = list(set(nodeS+nodeO)) # merge and uniq all nodes
  nodes = range(len(nodeAll)) # create node for the visualization (must be numbers)
  colors = ['maroon' if n in nodeS else 'TEAL' for n in nodeAll] 
  # set subject at color "maroon" and object as "teal" if both as "maroon"
  net = pyvis.network.Network(height='500px', width='1000px',) 
  # create a empty chart net
  net.add_nodes(nodes, label=nodeAll, color=colors, title=nodeAll) 
  # add nodes to network graph
  for i in svodf.index: # add relationships(Verbs) to the graph
    net.add_edge(nodeAll.index(svodf.loc[i,'Subject']),nodeAll.index(svodf.loc[i,'Object']),
     label=svodf.loc[i,'Verb'], title=str(i))
  net.show(filename) # save the graph to file "network.html"
  return filename

def docSVO(df, textCol='KeyTopic'):
  topicSVOs = {}
  for topic in df[textCol].unique():
    text = '\n'.join(df.loc[df[textCol]==topic,'content'])
    topicSVO = SVO(text)
    topicSVO.to_csv(f'svo_{topic}.csv', index=0)
    topicSVOs[topic] = topicSVO


if __name__ == '__main__':
  df = pd.read_excel('sampleData.xlsx')
  # df = df.sample(500).reset_index()
  textCol = 'content'
  nTopics = 5


  df['sentimentScore'] = df[textCol].apply(sentimentSocre)

  df['cleaned_text'] = df[textCol].apply(text_clean)

  doctopics, topicDescribe, ldaVis = topic_analysis(df['cleaned_text'], nTopics)
  df = df.merge(doctopics, left_index=True, right_index=True)

  topicSVOs = docSVO(df)

  visfile = visSVO(topicSVOs['topic2'],'topic2')
  IPython.display.HTML(visfile)

  plot_topics(topicDescribe)


#pip install IPython bokeh pandas xlrd gensim spacy textacy textblob pyvis pyLDAvis
# python -m spacy download en_core_web_sm

# from ztext.nlpsteps import text_clean, sentimentSocre, topic_analysis,plot_topics,svo

# import pandas as pd
import warnings,IPython, spacy
from tqdm import tqdm


warnings.filterwarnings('ignore')

def sampledata():
    import pandas as pd
    return pd.read_excel('https://github.com/ZackAnalysis/ztext/blob/master/ztext/sampleData.xlsx?raw=true')

    
class Ztext:
    def __init__(self,df,textCol,nTopics=5,custom_stopwrods=[], samplesize=None):
        try:
            nlp = spacy.load('en_core_web_sm')
        except:
            print('Please make sure spacy model is loaded.\n\
                    Install spacy model first by running\n\
                    python -m spacy download en_core_web_sm')
            return
        if samplesize:
            print('')
            df = df.sample(samplesize)
        self.df = df
        self.textCol = textCol
        self.nTopics = nTopics
        self.custom_stopwrods = custom_stopwrods
        self.doctopics = None
        self.topicDescribe = None
        self.dictionary = None
        self.model = None
        self.tokens = None
        self.ldaVis = None
        self.svodfs = {}
    
    def sentiment(self):
        print('sentment analyzing ...')
        from ztext.nlpsteps.sentimentSocre import sentimentSocre
        self.df['sentimentScore'] = self.df[self.textCol].apply(sentimentSocre)
        return self.df

    def clean(self):
        print('cleaning text ...')
        from ztext.nlpsteps.text_clean import text_clean
        for n in tqdm(self.df.index):
            self.df.loc[n,'cleaned_text'] = text_clean(self.df.loc[n,self.textCol])
        # self.df['cleaned_text'] = self.df[self.textCol].apply(text_clean, custom_stopwrods=self.custom_stopwrods)
        return self.df

    def get_topics(self, nTopics=None):
        
        if not nTopics:
            nTopics = self.nTopics
        from ztext.nlpsteps.topicAnalysis import topic_analysis
        if 'cleaned_text' not in self.df:
            print('data need to be cleaned first, cleaning the data')
            self.clean()
        print('getting topics ...')
        self.doctopics, self.topicDescribe, self.model, self.tokens, self.dictionary= topic_analysis(self.df, nTopics,'cleaned_text')
        self.df = self.df.merge(self.doctopics, left_index=True, right_index=True)
        return self.df

    def getldaVis(self):
        from ztext.nlpsteps.topicAnalysis import getldaVis
        
        if self.model is None or self.tokens is None:
            print('applying LDA analysis first')
            self.get_topics()
        self.ldaVis = getldaVis(self.model, self.tokens)
        return self.ldaVis

    def topicCount(self):
        from ztext.nlpsteps.topicAnalysis import plot_topics
        if self.topicDescribe is None:
            print('applying LDA analysis first')
            self.topic_analysis()
        plot_topics(self.topic_analysis)
        # to do add none notebook here


    def getSVO(self, topicN='topic1',clean=False):
        from ztext.nlpsteps.svo import SVO  
        if 'KeyTopic' not in self.df:
            print('Topic analysis must run first. ')
            self.get_topics()
        print('getting SVO for ', topicN)
        text = '\n'.join(self.df.loc[self.df['KeyTopic']==topicN,self.textCol])
        svodf = SVO(text)
        self.svodfs[topicN] = svodf
        return svodf
         
    
    def SVOall(self, topicCol='KeyTopic',clean=False):
        if clean:
            self.svodfs = {}
        from ztext.nlpsteps.svo import SVO  
        if 'KeyTopic' not in self.df:
            print('Topic analysis must run first. ')
            self.get_topics()
        print('Starting SVO extraction for all topics...')
        for topic in self.df[topicCol].unique():
            if topic in self.svodfs:
                print('SVO dateframe for ', topic, 'already exists, next.')
                continue
            print('analyzing ', topic, '...')
            self.getSVO(topic)
        return self.svodfs
    
    def getSVOvis(self, topic='topic1',options='entity'): # option could be "entity", "person", "any"
        from ztext.nlpsteps.svo import visSVO
        if topic not in self.svodfs:
            print('Must run SVO extraction first')
            self.getSVO(topic)
        if self.svodfs[topic] is None or len(self.svodfs[topic]) == 0:
            print('no entity detected in the corpus, try another topic')
            return
        print('creating SVO Visualization file')
        visSVO(self.svodfs[topic],topic)
        return

    def getAllentities(self):
        import pandas as pd
        SVOall()
        return pd.concat([svodfs.values()])


# if __name__ == '__main__':


    # import pandas as pd
    # df = pd.read_excel('ztext/ztext/sampleData.xlsx')
    # nlpdf = Nlpdf(df,'content',samplesize=300)
    # nlpdf.get_top




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="ztext", # Replace with your own username
    version="0.0.6",
    author="Zack Dai",
    author_email="zdai@brocku.ca",
    description="An easy to use NLP tool",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ZackAnalysis/ztext",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=setuptools.find_packages(),
    install_requires=["IPython","bokeh","pandas","xlrd","gensim","spacy",
        "textacy","textblob","pyvis","pyLDAvis"],
    python_requires='>=3.6',
)


import spacy,warnings


warnings.filterwarnings('ignore')
try:
  nlp = spacy.load('en_core_web_sm')
except:
  print('Install spacy model first by running\npython -m spacy download en_core_web_sm')

def text_clean(text, custom_stopwrods=[], toLower=True):

  if toLower:
    text=text.lower()
  doc = nlp(text)
  return ' '.join([d.lemma_ for d in doc if \
    all([2<len(d.text)<20,d.is_alpha, not d.is_stop, d.lemma_ \
      not in custom_stopwrods])])
